/******************************************
MidiSense V1.0 firmware

MidiSense firmware is distributed under CC license. 
For more info on CC go to www.creativecommons.org
For more info on MidiSense go to www.ladyada.net/make/midisense

Creative Commons Deed
Attribution-NonCommercial-ShareAlike 2.5
You are free:
    * to copy, distribute, display, and perform the work
    * to make derivative works

Under the following conditions:
Attribution.
   You must attribute the work in the manner specified by the 
   author or licensor.
Noncommercial. 
   You may not use this work for commercial purposes.
Share Alike. 
   If you alter, transform, or build upon this work, you may 
   distribute the resulting work only under a license identical to this one.

    * For any reuse or distribution, you must make clear to others 
      the license terms of this work.
    * Any of these conditions can be waived if you get permission 
      from the copyright holder.

Your fair use and other rights are in no way affected by the above.

A more detailed version of this license is available at:
http://creativecommons.org/licenses/by-nc-sa/2.5/legalcode
******************************************/


#include <avr/io.h>
#include <avr/interrupt.h>
#include "util.h"
#include "midi.h"
#include "main.h"

#define PRESAMPLERATE 32

#define VERSION_MAJ 1
#define VERSION_MIN 2

volatile uint8_t midi_addr;
extern uint8_t midi_state;

extern uint8_t uart_in_use;

volatile uint8_t rawvalue = 0xFF;

#define NUM_SENSORS 6
uint32_t oversamplingbuff[NUM_SENSORS];
uint8_t samplingnum[NUM_SENSORS];
uint8_t oversamplerate[NUM_SENSORS];
uint8_t hysteresisflag[NUM_SENSORS];
uint8_t notevel[NUM_SENSORS];
uint16_t prevdata[NUM_SENSORS];
uint8_t last_sensor;

extern uint8_t sysex_buff_len;
extern uint8_t sysex_buff[SYSEX_MAX_LEN];

void ioinit(void) {
  int i;

  DDRD = _BV(5); // PD5 is output led
  DDRB = _BV(2); // PB2 is output led

  PORTB = _BV(0) | _BV(1);
  PORTD = _BV(6) | _BV(7);

#if (BOARDTYPE == ANADIGIO)
  // set directions for each sensor
  for (i=0; i<11; i++) {
    configsensoriotype(i);
  }
#endif
}

void configsensoriotype(uint8_t sensor) {
  int type = read_sensoriotype(sensor);
  if (type == ANALOG_INPUT) {
    setsensordirection(sensor, INPUT);
    enablesensorpullup(sensor, 0);
    // no pullups, all input
  } else if (type == RIBBON_INPUT) {
    // turn on pullups
    setsensordirection(sensor, INPUT);
    enablesensorpullup(sensor, 1);
  } else if (type == DIGITAL_INPUT) {
    // turn on pullups
    setsensordirection(sensor, INPUT);
    enablesensorpullup(sensor, 1);
  } else if (type == PWM_OUTPUT) {
    setsensordirection(sensor, OUTPUT);
    setsensoroutput(sensor, 0);
  }
}

void setsensoroutput(uint8_t sensor, uint8_t flag) {
  enablesensorpullup(sensor, flag);
}

uint8_t readsensorinput(uint8_t sensor) {
  if (sensor < 6) {
    return PINC & _BV(sensor);
  } else if (sensor == 6) {
    return PINB & _BV(5);
  } else if (sensor == 7) {
    return PINB & _BV(4);
  } else if (sensor < 11) {
    return PIND & _BV(sensor-6);
  }
  return 0;
}

void enablesensorpullup(uint8_t sensor, uint8_t flag) {
  if (sensor < 6) {
    if (flag)
      PORTC |= _BV(sensor);
    else
      PORTC &= ~_BV(sensor);
  } else if (sensor == 6) {
    if (flag)
      PORTB |= _BV(5);
    else
      PORTB &= ~_BV(5);
  } else if (sensor == 7) {
    if (flag) 
      PORTB |= _BV(4);
    else
      PORTB &= ~_BV(4);   
  } else if (sensor < 11) {
    if (flag)
      PORTD |= _BV(sensor-6);
    else
      PORTD &= ~_BV(sensor-6);
  }
}

void setsensordirection(uint8_t sensor, uint8_t dir) {
  if (sensor < 6) {
    if (dir)
      DDRC |= _BV(sensor);
    else
      DDRC &= ~_BV(sensor);
  } else if (sensor == 6) {
    if (dir)
      DDRB |= _BV(5);
    else
      DDRB &= ~_BV(5);
  } else if (sensor == 7) {
    if (dir)
      DDRB |= _BV(4);
    else
      DDRB &= ~_BV(4);
  } else if (sensor < 11) {
    if (dir)
      DDRD |= _BV(sensor-6);
    else
      DDRD &= ~_BV(sensor-6);
  }
}

#if (BOARDTYPE == RESISTIVE) || (BOARDTYPE == ANADIGIO)

void adc_init(void) {
  LEDTX_PORT |= _BV(LEDTX);
  // enable ADC, enable interrupt, and use 64 divider = 125khz clock
  // 125Khz clock -> 8Khz sampling freq (8000 samples per second)
  ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS0);

  ADCSRA |= _BV(ADSC);  // start a conversion now to get the analog system up
  loop_until_bit_is_clear(ADCSRA, ADSC); // OK first conversion is done!
  
  // set up the interrupt that will do the sampling for us, at 8khz
  // we'll use timer 0 because we only need 8 bits resolution anyhow
  TCCR0 = _BV(CS01);  // clk div by 8
  TCNT0 = 130;                    // then we'll initialize w/200
  // 8Mhz / 8 = 1000Khz then if you count from 130 to 255 @ 1000Khz -> 8Khz
  TIMSK |= _BV(TOIE0);
}

// this gets called at 8Khz
uint32_t temp32 = 0;

SIGNAL(SIG_OVERFLOW0) {
  uint8_t curr_chan, t8;
  uint16_t data;
  uint16_t sensortype;
  uint16_t min, mid, max;

  TCNT0 = 130; // reset counter to keep 8khz timer

  if ( (ADCSRA & _BV(ADSC)) != 0 ) {
    // woah, why is a conversion still happening? try again later
    return;
  }
  
  // grab ADC data
  data = ADC;

  // go to the next channel & start next conversion!
  curr_chan = ADMUX & 0x7;
  if (curr_chan >= 5) {
    ADMUX = 0; // go back to first channel
  } else {
    ADMUX = curr_chan+1;
  }

#if (BOARDTYPE == ANADIGIO)
  // check if its an analog in
  t8 = read_sensoriotype(ADMUX);
  if ((t8 == ANALOG_INPUT) || (t8 == RIBBON_INPUT))
    ADCSRA |= _BV(ADSC);
  t8 = read_sensoriotype(curr_chan);
  if ((t8 != ANALOG_INPUT) && (t8 != RIBBON_INPUT))
    return;

  if (( read_sensoriotype(curr_chan) == RIBBON_INPUT) && (data >= 0x3F0)) {
    oversamplingbuff[curr_chan] = 0;
    samplingnum[curr_chan] = 0;
    return;
  } 


#else
  // start the ADC again
  ADCSRA |= _BV(ADSC);
#endif

  // meanwhile, keep the data logged....  
  oversamplingbuff[curr_chan] += data;
  // increase the oversample counter
  samplingnum[curr_chan]++;

  if (notevel[curr_chan] < 0x7F) {
    notevel[curr_chan]++;
  }

  if ( (samplingnum[curr_chan] / PRESAMPLERATE) > oversamplerate[curr_chan]) {
    // OK ready to send out the data!
    oversamplingbuff[curr_chan] /= PRESAMPLERATE; // first level of oversample
    // second level of oversample
    oversamplingbuff[curr_chan] /= (oversamplerate[curr_chan] + 1);

    if ((curr_chan < NUM_SENSORS) && 
	(read_sensorsenabled() & _BV(curr_chan))) {

      sensortype = read_sensortype(curr_chan);
      min = read_sensorcalib_min(curr_chan);
      mid = read_sensorcalib_mid(curr_chan);
      max = read_sensorcalib_max(curr_chan);

      // if its inverted, swap min & max
      if (min > max) {
	temp32 =  min;
	min = max;
	max = temp32;
      }

      temp32 = oversamplingbuff[curr_chan];
      //note_on(0xF, temp32 >> 7, temp32 & 0x7F);
 
      /* if we've had a request to send the raw value do it now! */
      if (rawvalue == curr_chan) {
	sysex_buff[0] = 0x69; // to computer 
	sysex_buff[1] = MSG_RAWVALUE;
	sysex_buff[2] = read_midiaddr() + 1; // the midiaddr
	sysex_buff[3] = curr_chan;           // the sensor num
	sysex_buff[4] = temp32 >> 3;
	sysex_buff[5] = (temp32 & 0x7) << 4;
	sysex_buff_len = 6;
	midi_send_sysex();                     
      }

      // if its note on/off we use the calibration data as a 'schmitt trigger'
      if (temp32 < mid) {
	hysteresisflag[curr_chan] |= _BV(0);
      } else if (temp32 > mid) {
	hysteresisflag[curr_chan] |= _BV(1);
      }

      if ( ((prevdata[curr_chan] < mid) && (temp32 > mid)) ||
	   ((prevdata[curr_chan] > mid) && (temp32 < mid)) )
	notevel[curr_chan] = 1;

      if ( ((sensortype>>8) == 0x7F) && 
	   ((((sensortype & 0xFF)|0x80) == MIDI_NOTEON) ||
	    (((sensortype & 0xFF)|0x80) == MIDI_START) ||
	    (((sensortype & 0xFF)|0x80) == MIDI_CONTINUE))) {
	if ((temp32 < min) && 
	    (hysteresisflag[curr_chan] & _BV(1)) &&
	    (last_sensor & _BV(curr_chan))) {

	  if (read_sensorcalib_min(curr_chan) < read_sensorcalib_max(curr_chan))
	    switch ((sensortype & 0xFF) | 0x80) {
	    case MIDI_NOTEON:	
	      //note_off(curr_chan, (read_sensorvars(curr_chan) >> 8), notevel[curr_chan]);
#if (VERSION_MIN == 1)
	      note_off(curr_chan, (read_sensorvars(curr_chan) >> 8), 0x64);
#else
	      note_off(midi_addr, (read_sensorvars(curr_chan) >> 8), 0x64);
#endif
	      notevel[curr_chan] = 0x80;
	      break;
	    case MIDI_START:
	    case MIDI_CONTINUE:
	      uart_putchar(MIDI_STOP);
	      break;
	    }	      
	  else
	    switch ((sensortype & 0xFF) | 0x80) {
	    case MIDI_NOTEON:
	      //note_on(curr_chan, (read_sensorvars(curr_chan) >> 8), notevel[curr_chan]);
#if (VERSION_MIN == 1)
	      note_on(curr_chan, (read_sensorvars(curr_chan) >> 8), 0x64);
#else
	      note_on(midi_addr, (read_sensorvars(curr_chan) >> 8), 0x64);
#endif
	      notevel[curr_chan] = 0x80;
	      break;
	    case MIDI_START:
	      uart_putchar(MIDI_START);
	      break;
	    case MIDI_CONTINUE:
	      uart_putchar(MIDI_CONTINUE);
	      break;
	    }	      

	  hysteresisflag[curr_chan] &= ~_BV(1);
	  last_sensor &= ~_BV(curr_chan);
	} else if ((temp32 > max) && 
		   (hysteresisflag[curr_chan] & _BV(0)) &&
		   !(last_sensor & _BV(curr_chan))) {

	  if (read_sensorcalib_min(curr_chan) < read_sensorcalib_max(curr_chan)) {
	    switch ((sensortype & 0xFF) | 0x80) {
	    case MIDI_NOTEON:
#if (VERSION_MIN == 1)
	      note_on(curr_chan, (read_sensorvars(curr_chan) >> 8), 0x64);
#else
	      note_on(midi_addr, (read_sensorvars(curr_chan) >> 8), 0x64);
#endif
	      notevel[curr_chan] = 0x80;
	      break;
	    case MIDI_START:
	      uart_putchar(MIDI_START);
	      break;
	    case MIDI_CONTINUE:
	      uart_putchar(MIDI_CONTINUE);
	      break;
	    }
	  } else {
	    switch ((sensortype & 0xFF) | 0x80) {
	    case MIDI_NOTEON:
	      //note_off(curr_chan, (read_sensorvars(curr_chan) >> 8), notevel[curr_chan]);
#if (VERSION_MIN == 1)
	      note_off(curr_chan, (read_sensorvars(curr_chan) >> 8), 0x64);
#else
	      note_off(midi_addr, (read_sensorvars(curr_chan) >> 8), 0x64);
#endif
	      notevel[curr_chan] = 0x80;
	      break;
	    case MIDI_START:
	    case MIDI_CONTINUE:
	      uart_putchar(MIDI_STOP);
	      break;
	    }	 
	  }
	  last_sensor |= _BV(curr_chan);
	  hysteresisflag[curr_chan] &= ~_BV(0);
	}
      } else {
	// normalize it by subtracting the minimum
	if (temp32 < min)
	  temp32 = 0;
	else
	  temp32 -=  min;
	// now multiply it by the ratio between the max and 1024
	temp32 *= 1024;
	temp32 /= max - min;
	// make sure we dont go above max by assigning to 16 bit val
	if (temp32 > 1023)
	  temp32 = 1023;
	data = temp32;
	
	if (read_sensorcalib_min(curr_chan) >
	    read_sensorcalib_max(curr_chan)) {
	  // inverty
	  data = 1023 - data;
	}
	
	if ((sensortype>>8) == 0x7F) {
	  switch ((sensortype & 0xFF) | 0x80) 
	    {
	    case MIDI_AFTERTOUCH:
	      // two databytes, 1st is note (var), 2nd is pressure
#if (VERSION_MIN == 1)
	      uart_putchar(sensortype | 0x80 | curr_chan);
#else
	      uart_putchar(sensortype | 0x80 | midi_addr);
#endif
	      uart_putchar((read_sensorvars(curr_chan) >> 8) & 0x7F);
	      uart_putchar(data >> 3);
	      break;
	      
	    case MIDI_CHANPRESSURE:
#if (VERSION_MIN == 1)
	      uart_putchar(sensortype | 0x80 | curr_chan);
#else
	      uart_putchar(sensortype | 0x80 | midi_addr);
#endif
	      uart_putchar(data >> 3);
	      break;
	      
	    case MIDI_PITCHWHEEL:
#if (VERSION_MIN == 1)
	      uart_putchar(sensortype | 0x80 | curr_chan);
#else
	      uart_putchar(sensortype | 0x80 | midi_addr);
#endif
	      uart_putchar((data & 0x7) << 4);
	      uart_putchar(data >> 3);
	      break;
	      
	    }
	} else {
	  // its a controller message! this is different than above
#if (VERSION_MIN == 1)	  
	  sendmidicc(curr_chan, sensortype >> 8, data >> 3);
#else
	  sendmidicc(midi_addr, sensortype >> 8, data >> 3);
#endif
	  // if there's a 'fine' setting, send that too
	  if ((sensortype & 0xFF) != 0x7F) {
#if (VERSION_MIN == 1)
	    sendmidicc(curr_chan, sensortype & 0xFF, (data & 0x7) << 4);
#else
	    sendmidicc(midi_addr, sensortype & 0xFF, (data & 0x7) << 4);
#endif
	  }
	}
      }

    }
    
    // ok reset it too
    prevdata[curr_chan] = oversamplingbuff[curr_chan];
    oversamplingbuff[curr_chan] = 0;
    samplingnum[curr_chan] = 0;
    //note_off(0,88,66);
  }

}

#endif

volatile uint8_t ledtx_timeout = 0;
volatile uint8_t ledrx_timeout = 0;

SIGNAL(SIG_OVERFLOW2) {
  if (ledrx_timeout != 0) {
    ledrx_timeout--;
  } else {
    LEDRX_PORT &= ~_BV(LEDRX);
  }

  if (ledtx_timeout != 0) {
    ledtx_timeout--;
  } else {
    LEDTX_PORT &= ~_BV(LEDTX);
  }
}

uint16_t lastswitchreading;

int main(void) {
  uint8_t i, t8;
  uint8_t c;
  uint16_t t16, sensortype;

  ioinit();
  uart_init(BRRL_MIDI);
  midi_addr = read_midiaddr();
  
  // 8Mhz/32/256 = ~1Khz
  TCCR2 = _BV(CS21) | _BV(CS20);  // clk div by 32
  TIMSK |= _BV(TOIE2);

#if (BOARDTYPE == RESISTIVE) || (BOARDTYPE == ANADIGIO)
  // set up sampling interrupts!
  adc_init();
  // set up sampling data
  for (i=0; i<6; i++) {
    samplingnum[i] = 0;
    oversamplingbuff[i] = 0;
    oversamplerate[i] = read_sensorrate(i);
  }

#elif (BOARDTYPE == QPROXTOUCH)
  qprox_init();
#endif

  // turn on interrupts and start spitting out data!
  sei();

  while (1) {
#if (BOARDTYPE == ANADIGIO)
    // if we have digital switch inputs, check them
    
    // first see if they are enabled 
    t16 = read_sensorsenabled();
    for (i=0; i<NUMSENSORS; i++) {
      // 5 digital sensors
      if (! (t16 & _BV(i)))
	continue; // not enabled
      if (read_sensoriotype(i) != DIGITAL_INPUT)
	continue; // not a switch input

      continue;

      t8 = readsensorinput(i);
      if ( (lastswitchreading & _BV(i)) && !t8 ) {
	// note turned off
	c = 0;
	lastswitchreading &= ~_BV(i);
      } else if ( !(lastswitchreading & _BV(i)) && t8) {
	// note turned on
	c = 0x7F;
	lastswitchreading |= _BV(i);
      } else {
	continue;
      }

      sensortype = read_sensortype(i);
      if ((sensortype>>8) == 0x7F) {
	switch ((sensortype & 0xFF) | 0x80) 
	  {
	  case MIDI_AFTERTOUCH:
	    // two databytes, 1st is note (var), 2nd is pressure
	    cli();
	    uart_putchar(sensortype | 0x80 | midi_addr);
	    uart_putchar((read_sensorvars(i) >> 8) & 0x7F);
	    uart_putchar(c);
	    sei();
	    break;
	  case MIDI_CHANPRESSURE:
	    cli();
	    uart_putchar(sensortype | 0x80 | midi_addr);
	    uart_putchar(c);
	    sei();
	    break;
	  case MIDI_PITCHWHEEL:
	    cli();
	    uart_putchar(sensortype | 0x80 | midi_addr);
	    uart_putchar(0x22);
	    uart_putchar(0x22);
	    sei();
	    break;
	  }
      } else {
	// its a controller message! this is different than above
	sendmidicc(midi_addr, sensortype >> 8, 0x88);
	// if there's a 'fine' setting, send that too
	if ((sensortype & 0xFF) != 0x7F) {
	  sendmidicc(midi_addr, sensortype & 0xFF, c);
	}
      }
    }

#endif

    if (midi_getch()) {
      c = midi_getchar();
      //uart_putchar(c); // pass on to MIDI THRU
      if ((c & 0x80) != 0) { // command!
	if (c == MIDI_SYSEX_START) {
	  sysex_buff_len = 0;
	  midi_state = MIDI_READSYSEX;
	} else if (c == MIDI_SYSEX_FINISH) {
	  midi_state = MIDI_IDLE;
	} else {
	  // abort sysex 
	  sysex_buff_len = 0;
	  midi_state = MIDI_IDLE;
	}
      } else { // data byte
	if (midi_state == MIDI_READSYSEX) {
	  sysex_buff[sysex_buff_len++] = c;
	}
      }
    }

    if (midi_state == MIDI_IDLE) {
      if (sysex_buff_len != 0) { // a letter?? for me???
	midi_handle_sysex();
      }
    }
  }
}

void midi_handle_sysex(void) {
  uint8_t msgtype, crc, i;
  uint16_t t16;

  crc = sysex_buff[sysex_buff_len-1];
  sysex_buff_len--;

  if (sysex_buff_len < 2)
    return;

  // check th' CRC!

  if ((crc != 0) && (crc != calc_CRC7(sysex_buff, sysex_buff_len))) {
    for (i=0; i<sysex_buff_len; i++) {
      //midi_send_sysex_num(sysex_buff[i]);
      //midi_send_sysex_num(calc_CRC7(sysex_buff+i, 1));
    }
    return;
  }

  // check if the message is for us!
  if (sysex_buff[0] != MIDI_MANUF_ID) {
    sysex_buff_len = 0;
    return;
  }

  msgtype = sysex_buff[1];
  if ((msgtype != MSG_PING) && (sysex_buff[2] != read_midiaddr()+1))
    return;

  switch (msgtype) {
  case MSG_PING:
    // ok respond back
    sysex_buff[0] = 0x69; // to computer 
    sysex_buff[1] = MSG_PONG;
    sysex_buff[2] = read_midiaddr() + 1;
    sysex_buff[3] = BOARDTYPE;     // send the type
    sysex_buff[4] = VERSION_MAJ;   // send the firmware rev
    sysex_buff[5] = VERSION_MIN;
    sysex_buff_len = 6;
    midi_send_sysex();
    break;

  case MSG_SETSENSORIOTYPE:
    write_sensoriotype(sysex_buff[3], sysex_buff[4]);
    configsensoriotype(sysex_buff[3]);
  case MSG_REQSENSORIOTYPE:
    sysex_buff[0] = 0x69;
    sysex_buff[1] = MSG_SENSORIOTYPE;
    sysex_buff[4] = read_sensoriotype(sysex_buff[3]);
    sysex_buff_len = 5;
    midi_send_sysex();
    break;

  case MSG_SETSENSETYPE:
    write_sensortype(sysex_buff[3]-1, (sysex_buff[4] << 8) | sysex_buff[5]);
  case MSG_REQSENSETYPE:
    sysex_buff[0] = 0x69;
    sysex_buff[1] = MSG_SENSETYPE;
    t16 = read_sensortype(sysex_buff[3]-1);
    sysex_buff[4] = t16 >> 8;
    sysex_buff[5] = t16 & 0xFF;
    sysex_buff_len = 6;
    midi_send_sysex();
    break;

  case MSG_SETSENSORSENABLED:
    write_sensorsenabled(sysex_buff[3]);
  case MSG_REQSENSORSENABLED:
    sysex_buff[0] = 0x69;
    sysex_buff[1] = MSG_SENSORSENABLED;
    sysex_buff[2] = read_midiaddr() + 1;
    t16 = read_sensorsenabled();
    sysex_buff[3] = (t16 >> 7) & 0x7F;
    sysex_buff[4] = t16 & 0x7F;
    sysex_buff_len = 5;
    midi_send_sysex();
    break;
    
  case MSG_RAWVALUESTART: {
   if (read_midiaddr() + 1 != sysex_buff[2])
     break;
   rawvalue = sysex_buff[3];
   break;
  }
  case MSG_RAWVALUESTOP: {
    if (read_midiaddr() + 1 != sysex_buff[2])
      break;
    rawvalue = 0xFF;
    break;
  }

  case MSG_SETSENSORCALIB:
    write_sensorcalib(sysex_buff[3], // sensornum
		      ((uint16_t)sysex_buff[4] << 7) | sysex_buff[5],
		      ((uint16_t)sysex_buff[6] << 7) | sysex_buff[7],
		      ((uint16_t)sysex_buff[8] << 7) | sysex_buff[9]);
  case MSG_REQSENSORCALIB:
    sysex_buff[0] = 0x69;
    sysex_buff[1] = MSG_SENSORCALIB;
    sysex_buff[2] = read_midiaddr() + 1;
    t16 = read_sensorcalib_min(sysex_buff[3]);
    sysex_buff[4] = t16 >> 7;
    sysex_buff[5] = t16 & 0x7F;
    t16 = read_sensorcalib_mid(sysex_buff[3]);
    sysex_buff[6] = t16 >> 7;
    sysex_buff[7] = t16 & 0x7F;
    t16 = read_sensorcalib_max(sysex_buff[3]);
    sysex_buff[8] = t16 >> 7;
    sysex_buff[9] = t16 & 0x7F;

    sysex_buff_len = 10;
    midi_send_sysex();
    break;  

  case MSG_SETSENSORVAR:
    write_sensorvars(sysex_buff[3], // sensornum
		     (sysex_buff[4] << 8)|sysex_buff[5] ); // sensor variable
  case MSG_REQSENSORVAR:
    sysex_buff[0] = 0x69;
    sysex_buff[1] = MSG_SENSORVAR;
    sysex_buff[2] = read_midiaddr() + 1;
    t16 = read_sensorvars(sysex_buff[3]);
    sysex_buff[4] = t16>>8;
    sysex_buff[5] = t16;
    sysex_buff_len = 6;
    midi_send_sysex();
    break;  

 case MSG_SETSENSORRATE:
    write_sensorrate(sysex_buff[3], // sensornum
		     sysex_buff[4]); // sensor variable
    oversamplerate[sysex_buff[3]] = read_sensorrate(sysex_buff[3]);
 case MSG_REQSENSORRATE:
   sysex_buff[0] = 0x69;
   sysex_buff[1] = MSG_SENSORRATE;
   sysex_buff[2] = read_midiaddr() + 1;
   sysex_buff[4] = read_sensorrate(sysex_buff[3]);
   sysex_buff_len = 5;
   midi_send_sysex();
   break;  
  }

}


uint16_t read_sensorcalib(uint8_t sensornum, uint8_t eeaddr) {
  if (sensornum >= NUMSENSORS)
    return 0;
  else
    return ((internal_eeprom_read(eeaddr + 2*sensornum) << 8) |
	    internal_eeprom_read(eeaddr + 2*sensornum + 1));
}
uint16_t read_sensorcalib_min(uint8_t sensornum) {
  return read_sensorcalib(sensornum, EE_CALIBRATED_MIN);
}
uint16_t read_sensorcalib_mid(uint8_t sensornum) {
  return read_sensorcalib(sensornum, EE_CALIBRATED_MID);
}
uint16_t read_sensorcalib_max(uint8_t sensornum) {
  return read_sensorcalib(sensornum, EE_CALIBRATED_MAX);
}


void write_sensorcalib(uint8_t sensornum, 
		       uint16_t min, uint16_t mid, uint16_t max) {
  if (sensornum >= NUMSENSORS)
    return;
  
  internal_eeprom_write(EE_CALIBRATED_MIN + 2*sensornum, min >> 8);
  internal_eeprom_write(EE_CALIBRATED_MIN + 2*sensornum + 1, min);
  
  internal_eeprom_write(EE_CALIBRATED_MID + 2*sensornum, mid >> 8);
  internal_eeprom_write(EE_CALIBRATED_MID + 2*sensornum + 1, mid);

  internal_eeprom_write(EE_CALIBRATED_MAX + 2*sensornum, max >> 8);
  internal_eeprom_write(EE_CALIBRATED_MAX + 2*sensornum + 1, max);
}

uint8_t read_sensorrate(uint8_t sensenum) {
  if (sensenum >= NUMSENSORS)
    return 0xFF;
  else
    return (internal_eeprom_read(EE_SENSORRATE + sensenum));
}

void write_sensorrate(uint8_t sensenum, uint8_t var) {
  if (sensenum >= NUMSENSORS)
    return;
  else
    internal_eeprom_write(EE_SENSORRATE + sensenum, var);
}

uint16_t read_sensorvars(uint8_t sensenum) {
  if (sensenum >= NUMSENSORS)
    return 0xFF;
  else
    return ((internal_eeprom_read(EE_SENSORVARS + 2*sensenum) << 8) |
	    internal_eeprom_read(EE_SENSORVARS + 2*sensenum + 1));
}

void write_sensorvars(uint8_t sensenum, uint16_t var) {
  if (sensenum >= NUMSENSORS)
    return;
  
  internal_eeprom_write(EE_SENSORVARS + 2*sensenum, var >> 8);
  internal_eeprom_write(EE_SENSORVARS + 2*sensenum+1, var);
}

uint8_t read_sensoriotype(uint8_t sensenum) {
  return (internal_eeprom_read(EE_IOTYPE+sensenum));
}

void write_sensoriotype(uint8_t sensenum, uint8_t type) {
  if (sensenum >= NUMSENSORS)
    return;
  internal_eeprom_write(EE_IOTYPE+sensenum, type);
}

uint16_t read_sensorsenabled(void) {
  return ((internal_eeprom_read(EE_SENSORSENABLED) << 8) |
	  internal_eeprom_read(EE_SENSORSENABLED + 1));
}

void write_sensorsenabled(uint16_t d) {
  internal_eeprom_write(EE_SENSORSENABLED, d>>8);
  internal_eeprom_write(EE_SENSORSENABLED+1, d);
}

uint16_t read_sensortype(uint8_t sensenum) {
  if (sensenum >= NUMSENSORS)
    return 0xFF;
  else
    return ((internal_eeprom_read(EE_SENSORTYPE_ADDR + 2*sensenum) << 8) |
	    internal_eeprom_read(EE_SENSORTYPE_ADDR + 2*sensenum + 1));
}

void write_sensortype(uint8_t sensenum, uint16_t type) {
  if (sensenum >= NUMSENSORS)
    return;
  
  internal_eeprom_write(EE_SENSORTYPE_ADDR + 2*sensenum, type >> 8);
  internal_eeprom_write(EE_SENSORTYPE_ADDR + 2*sensenum + 1, type);
}

