#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "util.h"
#include "main.h"
#include "midi.h"


const unsigned char CRC7Table[128] = {
    0x00,0x07,0x0E,0x09,0x1C,0x1B,0x12,0x15,
    0x38,0x3F,0x36,0x31,0x24,0x23,0x2A,0x2D,
    0x70,0x77,0x7E,0x79,0x6C,0x6B,0x62,0x65,
    0x48,0x4F,0x46,0x41,0x54,0x53,0x5A,0x5D,
    0x27,0x20,0x29,0x2E,0x3B,0x3C,0x35,0x32,
    0x1F,0x18,0x11,0x16,0x03,0x04,0x0D,0x0A,
    0x57,0x50,0x59,0x5E,0x4B,0x4C,0x45,0x42,
    0x6F,0x68,0x61,0x66,0x73,0x74,0x7D,0x7A,
    0x69,0x6E,0x67,0x60,0x75,0x72,0x7B,0x7C,
    0x51,0x56,0x5F,0x58,0x4D,0x4A,0x43,0x44,
    0x19,0x1E,0x17,0x10,0x05,0x02,0x0B,0x0C,
    0x21,0x26,0x2F,0x28,0x3D,0x3A,0x33,0x34,
    0x4E,0x49,0x40,0x47,0x52,0x55,0x5C,0x5B,
    0x76,0x71,0x78,0x7F,0x6A,0x6D,0x64,0x63,
    0x3E,0x39,0x30,0x37,0x22,0x25,0x2C,0x2B,
    0x06,0x01,0x08,0x0F,0x1A,0x1D,0x14,0x13};

uint8_t sysex_buff_len = 0;
uint8_t sysex_buff[SYSEX_MAX_LEN];

// flag to make sure only one thing is using the usart at once
volatile uint8_t uart_in_use = 0;

extern volatile uint8_t midi_addr;
uint8_t midi_state = MIDI_IDLE;
#define MIDI_Q_SIZE 64
volatile uint8_t midi_q[MIDI_Q_SIZE];      // cyclic queue for midi msgs
volatile uint8_t head_idx = 0;
volatile uint8_t tail_idx = 0;

extern volatile uint8_t ledrx_timeout;

// interrupt on receive char
SIGNAL(SIG_UART_RECV) {
  char c = UDR;

  LEDRX_PORT |= _BV(LEDRX);
  ledrx_timeout = 30;

  midi_q[tail_idx++] = c;    // place at end of q
  tail_idx %= MIDI_Q_SIZE;

  if (tail_idx == head_idx) {
    // i.e. there are too many msgs in the q
    // drop the oldest msg?
    head_idx++;
    head_idx %= MIDI_Q_SIZE;
  }
}

int midi_getch(void) {     // checks if there is a character waiting!
  if (head_idx != tail_idx)
    return 1;
  return 0;
}


int midi_getchar(void) {
  char c;

  LEDRX_PORT |= _BV(LEDRX);
  while (head_idx == tail_idx);
  LEDRX_PORT &= ~_BV(LEDRX);

  cli();
  c = midi_q[head_idx++];
  head_idx %= MIDI_Q_SIZE;
  sei();

  return c;
}

void midi_send_sysex_str(const char *str){ 
  char c;

  cli();

  uart_putchar(MIDI_SYSEX_START);
  while ((c = pgm_read_byte(str++)) != 0)
    uart_putchar(c);
  uart_putchar(MIDI_SYSEX_FINISH);

  sei();
}

void midi_send_sysex_num(uint8_t i) {
  cli();

  uart_putchar(MIDI_SYSEX_START);
  uart_putchar(0x69);
  uart_putchar(MSG_DEBUG);
  uart_putchar('0'+ i/100);
  i %= 100;
  uart_putchar('0'+ i/10);
  uart_putchar('0'+ i%10);
  uart_putchar(MIDI_SYSEX_FINISH);

  sei();
}


void midi_send_sysex(void) {
  uint8_t i;

  cli();

  uart_putchar(MIDI_SYSEX_START);
  for (i=0; i<sysex_buff_len; i++) 
    uart_putchar(sysex_buff[i] & 0x7F);
  // send the CRC
  uart_putchar(calc_CRC7(sysex_buff, sysex_buff_len));
  uart_putchar(MIDI_SYSEX_FINISH);

  sei();
}


void sendmidicc(uint8_t addr, uint8_t controller, uint8_t data) {
  cli();
  
  uart_putchar(MIDI_CC | addr);
  uart_putchar(controller & 0x7F);
  uart_putchar(data & 0x7F);
  
  sei();
}

uint8_t calc_CRC7(uint8_t *buff, uint8_t size) {
  uint8_t i;
  uint8_t crc = 0;
  
  /*
   * Add the message header to the CRC.  Don't include the CRC itself
   * when calculating the CRC.
   */
  for (i=0; i < size; i++) {	
    crc = CRC7Table[crc ^ *(buff+i)];
  }
  
  return crc;
}

void note_on(uint8_t addr, uint8_t note, uint8_t vel) {
  cli();

  uart_putchar(MIDI_NOTEON | addr);
  uart_putchar(note & 0x7F);
  uart_putchar(vel & 0x7F);

  sei();
}

void note_off(uint8_t addr, uint8_t note, uint8_t vel) {
  cli();

  uart_putchar(MIDI_NOTEOFF | addr);
  uart_putchar(note & 0x7F);
  uart_putchar(vel & 0x7F);

  sei();
}

uint8_t read_midiaddr(void) {
  return (((PINB & _BV(1)) << 2) | ((PINB & _BV(0)) << 2) 
	  | ((PIND & _BV(7)) >> 6) | ((PIND & _BV(6)) >> 6) );
}

